//========================================================================================
//  xeijwin.c
//  Copyright (C) 2003-2022 Makoto Kamada
//
//  This file is part of the XEiJ (X68000 Emulator in Java).
//  You can use, modify and redistribute the XEiJ if the conditions are met.
//  Read the XEiJ License for more details.
//  https://stdkmd.net/xeij/
//========================================================================================

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <windows.h>
#include <jni.h>

#include "xeij_NamedPipeInputStream.h"
#include "xeij_NamedPipeInputStream_Win.h"

#define VERSION 20221015

//========================================================================

void throwIndexOutOfBoundsException (JNIEnv *env, char *mes);
void throwIOException (JNIEnv *env, char *mes);
void throwNullPointerException (JNIEnv *env, char *mes);

//------------------------------------------------------------------------
//void throwIndexOutOfBoundsException (env, mes)
//  IndexOutOfBoundsExceptionをthrowする
//  ThrowNewは直ちにアボートしないのでreturnで明示的に中断させること
void throwIndexOutOfBoundsException (JNIEnv *env, char *mes) {
  (*env)->ThrowNew (env, (*env)->FindClass (env, "java/lang/IndexOutOfBoundsException"), mes);
}

//------------------------------------------------------------------------
//void throwIOException (env, mes)
//  IOExceptionをthrowする
//  ThrowNewは直ちにアボートしないのでreturnで明示的に中断させること
void throwIOException (JNIEnv *env, char *mes) {
  (*env)->ThrowNew (env, (*env)->FindClass (env, "java/io/IOException"), mes);
}

//------------------------------------------------------------------------
//void throwNullPointerException (env, mes)
//  NullPointerExceptionをthrowする
//  ThrowNewは直ちにアボートしないのでreturnで明示的に中断させること
void throwNullPointerException (JNIEnv *env, char *mes) {
  (*env)->ThrowNew (env, (*env)->FindClass (env, "java/lang/NullPointerException"), mes);
}

//========================================================================
//  NamedPipeInputStream
//========================================================================

//------------------------------------------------------------------------
//v = version ()
//  バージョンを返す
JNIEXPORT jint JNICALL Java_xeij_NamedPipeInputStream_version (JNIEnv *env, jobject obj) {
  return VERSION;
}

//========================================================================
//  NamedPipeInputStream.Win
//========================================================================

#define BUFFER_SIZE 8192

typedef struct {
  char *path;  //名前付きパイプのパス。プレフィックス"\\\\.\\pipe\\"を含む
  HANDLE handle;  //ハンドル
} workarea_t;

//------------------------------------------------------------------------
//length = available ()
//  ブロックせずに入力できる長さを返す
JNIEXPORT jint JNICALL Java_xeij_NamedPipeInputStream_00024Win_available (JNIEnv *env, jobject obj) {
  //変数を取り出す
  jclass cls = (*env)->GetObjectClass (env, obj);
  jfieldID wpid = (*env)->GetFieldID (env, cls, "wp", "J");
  workarea_t *wp = (workarea_t *) (*env)->GetLongField (env, obj, wpid);
  //閉じていたらエラー
  if (wp == NULL || wp->handle == INVALID_HANDLE_VALUE) {
    throwIOException (env, "named pipe not open");
    return 0;
  }
  //ブロックせずに入力できる長さを求める
  jint t = 0;
  //  https://learn.microsoft.com/ja-jp/windows/win32/api/namedpipeapi/nf-namedpipeapi-peeknamedpipe
  if (!PeekNamedPipe (wp->handle, NULL, 0, NULL, (LPDWORD) &t, NULL)) {
    //  https://learn.microsoft.com/en-us/windows/win32/debug/system-error-codes
    jint error = GetLastError ();
    if (error == ERROR_BAD_PIPE ||  //230 パイプの状態が無効。誰も開いていない
        error == ERROR_BROKEN_PIPE) {  //109 パイプが終了した。誰かが開いて閉じた
      t = 0;
    } else {
      char mes[256];
      sprintf (mes, "PeekNamedPipe returned error code %ld", error);
      throwIOException (env, mes);
      return -1;
    }
  }
  //ブロックせずに入力できる長さを返す
  return t;
}

//------------------------------------------------------------------------
//cancel ()
//  操作を取り消す
JNIEXPORT void JNICALL Java_xeij_NamedPipeInputStream_00024Win_cancel (JNIEnv *env, jobject obj) {
  //変数を取り出す
  jclass cls = (*env)->GetObjectClass (env, obj);
  jfieldID wpid = (*env)->GetFieldID (env, cls, "wp", "J");
  workarea_t *wp = (workarea_t *) (*env)->GetLongField (env, obj, wpid);
  //閉じていたら何もしない
  if (wp == NULL) {
    return;
  }
  HANDLE handle = wp->handle;
  if (handle == INVALID_HANDLE_VALUE) {
    return;
  }
  //操作を取り消す
  //  CancelIoExで取り消された操作はERROR_OPERATION_ABORTEDが返る
  //  https://learn.microsoft.com/ja-jp/windows/win32/fileio/cancelioex-func
  if (!CancelIoEx (handle, NULL)) {
    //jint error = GetLastError ();
    //if (error == ERROR_NOT_FOUND) {  //1168 取り消す操作がない
    //}
    //char mes[256];
    //sprintf (mes, "CancelIoEx returned error code %ld", error);
  }
}

//------------------------------------------------------------------------
//close ()
//  名前付きパイプを閉じる
JNIEXPORT void JNICALL Java_xeij_NamedPipeInputStream_00024Win_close (JNIEnv *env, jobject obj) {
  //変数を取り出す
  jclass cls = (*env)->GetObjectClass (env, obj);
  jfieldID wpid = (*env)->GetFieldID (env, cls, "wp", "J");
  workarea_t *wp = (workarea_t *) (*env)->GetLongField (env, obj, wpid);
  //閉じていたら何もしない
  if (wp == NULL || wp->handle == INVALID_HANDLE_VALUE) {
    return;
  }
  //名前付きパイプを閉じる
  //  https://learn.microsoft.com/ja-jp/windows/win32/api/handleapi/nf-handleapi-closehandle
  if (!CloseHandle (wp->handle)) {
    //jint error = GetLastError ();
    //char mes[256];
    //sprintf (mes, "CloseHandle returned error code %ld", error);
    //wp->handle = INVALID_HANDLE_VALUE;
  }
  //wp->handle = INVALID_HANDLE_VALUE;
  free (wp->path);
  free (wp);
  wp = NULL;
  //変数を書き戻す
  (*env)->SetLongField (env, obj, wpid, (__int64) wp);
}

//------------------------------------------------------------------------
//connect ()
//  接続を待つ
JNIEXPORT void JNICALL Java_xeij_NamedPipeInputStream_00024Win_connect (JNIEnv *env, jobject obj) {
  //変数を取り出す
  jclass cls = (*env)->GetObjectClass (env, obj);
  jfieldID wpid = (*env)->GetFieldID (env, cls, "wp", "J");
  workarea_t *wp = (workarea_t *) (*env)->GetLongField (env, obj, wpid);
  //閉じていたらエラー
  if (wp == NULL || wp->handle == INVALID_HANDLE_VALUE) {
    throwIOException (env, "named pipe not open");
    return;
  }
  //接続を待つ
  //  https://learn.microsoft.com/ja-jp/windows/win32/api/namedpipeapi/nf-namedpipeapi-connectnamedpipe
  if (!ConnectNamedPipe (wp->handle, NULL)) {
    jint error = GetLastError ();
    if (error == ERROR_PIPE_CONNECTED ||  //535 既に接続している
        error == ERROR_NO_DATA ||  //232 既に切断された
        error == ERROR_OPERATION_ABORTED) {  //995 スレッドが終了したか操作が取り消された
    } else {
      char mes[256];
      sprintf (mes, "ConnectNamedPipe returned error code %ld", error);
      throwIOException (env, mes);
      return;
    }
  }
}

//------------------------------------------------------------------------
//open (name)
//  名前付きパイプを開く
//  name  名前付きパイプの名前。プレフィックス"\\\\.\\pipe\\"を含まない
JNIEXPORT void JNICALL Java_xeij_NamedPipeInputStream_00024Win_open (JNIEnv *env, jobject obj, jstring name) {
  //ワークエリアを確保する
  workarea_t *wp = malloc (sizeof (workarea_t));
  //名前付きパイプのパスを作る
  const char *utf = (*env)->GetStringUTFChars (env, name, NULL);
  jint strlen_utf = strlen (utf);
  wp->path = malloc (9 + strlen_utf + 1);
  strcpy_s (wp->path, 9 + strlen_utf + 1, "\\\\.\\pipe\\");
  strcpy_s (wp->path + 9, strlen_utf + 1, utf);
  (*env)->ReleaseStringUTFChars (env, name, utf);
  //名前付きパイプを開く
  //  https://learn.microsoft.com/ja-jp/windows/win32/api/winbase/nf-winbase-createnamedpipea
  wp->handle = CreateNamedPipeA (wp->path,  //LPCSTR lpName
                                 PIPE_ACCESS_INBOUND,  //DWORD dwOpenMode
                                 PIPE_TYPE_BYTE | PIPE_WAIT,  //DWORD dwPipeMode
                                 PIPE_UNLIMITED_INSTANCES,  //DWORD nMaxInstances
                                 0,  //DWORD nOutBufferSize
                                 BUFFER_SIZE,  //DWORD nInBufferSize
                                 0,  //DWORD nDefaultTimeOut
                                 NULL);  //LPSECURITY_ATTRIBUTES lpSecurityAttributes
  //開けなかったらエラー
  if (wp->handle == INVALID_HANDLE_VALUE) {
    throwIOException (env, "CreateNamedPipeA failed");
    return;
  }
  //変数を書き戻す
  jclass cls = (*env)->GetObjectClass (env, obj);
  jfieldID wpid = (*env)->GetFieldID (env, cls, "wp", "J");
  (*env)->SetLongField (env, obj, wpid, (__int64) wp);
}

//------------------------------------------------------------------------
//data = read ()
//  1バイト入力して入力したデータを返す
//  1バイト入力するまでブロックする
//  名前付きパイプにはEOFがないので-1が返ることはない
//  data  入力したデータ。0-255
JNIEXPORT jint JNICALL Java_xeij_NamedPipeInputStream_00024Win_read__ (JNIEnv *env , jobject obj) {
  //変数を取り出す
  jclass cls = (*env)->GetObjectClass (env, obj);
  jfieldID wpid = (*env)->GetFieldID (env, cls, "wp", "J");
  workarea_t *wp = (workarea_t *) (*env)->GetLongField (env, obj, wpid);
  //閉じていたらエラー
  if (wp == NULL || wp->handle == INVALID_HANDLE_VALUE) {
    throwIOException (env, "named pipe not open");
    return -1;
  }
  //1バイト入力する
  char buf[1];
  buf[0] = 0;
  jint off = 0;
  jint len = 1;
  jint k = 0;
  while (k < len) {
    jint t = 0;
    //  https://learn.microsoft.com/ja-jp/windows/win32/api/fileapi/nf-fileapi-readfile
    if (!ReadFile (wp->handle, &buf[off + k], len - k, (LPDWORD) &t, NULL)) {
      jint error = GetLastError ();
      if (error == ERROR_BROKEN_PIPE ||  //109 パイプが終了した。誰かが開いて閉じた
          error == ERROR_PIPE_LISTENING || //536 誰も開いていない
          error == ERROR_OPERATION_ABORTED) {  //995 スレッドが終了したか操作が取り消された
        break;
      } else {
        char mes[256];
        sprintf (mes, "ReadFile returned error code %ld", error);
        throwIOException (env, mes);
        return 0;
      }
    }
    k += t;
  }
  //入力したデータを返す
  return buf[0] & 0xff;
}

//------------------------------------------------------------------------
//len = read (b)
//  配列に入力して入力した長さを返す
//  配列の長さが0のときは何もせず0を返す。さもなくば少なくとも1バイト入力するまでブロックする
//  len  入力した長さ
//  b    配列
JNIEXPORT jint JNICALL Java_xeij_NamedPipeInputStream_00024Win_read___3B (JNIEnv *env, jobject obj, jbyteArray b) {
  //read(b,0,b.length)を実行する
  return Java_xeij_NamedPipeInputStream_00024Win_read___3BII (env, obj, b, 0, (*env)->GetArrayLength (env, b));
}

//------------------------------------------------------------------------
//len = read (b, off, len)
//  配列に入力して入力した長さを返す
//  入力する長さが0のときは何もせず0を返す。さもなくば少なくとも1バイト入力するまでブロックする
//  len  入力した長さ
//  b    配列
//  off  開始位置
//  len  入力する長さ
JNIEXPORT jint JNICALL Java_xeij_NamedPipeInputStream_00024Win_read___3BII (JNIEnv *env, jobject obj, jbyteArray b, jint off, jint len) {
  //変数を取り出す
  jclass cls = (*env)->GetObjectClass (env, obj);
  jfieldID wpid = (*env)->GetFieldID (env, cls, "wp", "J");
  workarea_t *wp = (workarea_t *) (*env)->GetLongField (env, obj, wpid);
  //閉じていたらエラー
  if (wp == NULL || wp->handle == INVALID_HANDLE_VALUE) {
    throwIOException (env, "named pipe not open");
    return -1;
  }
  //開始位置と入力する長さを確認する
  jint cap = (*env)->GetArrayLength (env, b);  //配列の長さ
  if (off < 0 || len < 0 || cap < off + len) {  //範囲外
    char mes[256];
    sprintf (mes, "b.length=%ld, off=%ld, len=%ld", cap, off, len);
    throwIndexOutOfBoundsException (env, mes);
    return 0;
  }
  //入力する長さが0のときは何もせず0を返す
  if (len == 0) {
    return 0;
  }
  //lenバイト入力する
  //  GetPrimitiveArrayCriticalに成功したらReleasePrimitiveArrayCriticalまで他のJNI関数を呼び出してはならない。入れ子にすることはできる
  char *buf = (char *) (*env)->GetPrimitiveArrayCritical (env, b, 0);
  jint k = 0;
  while (k < 1) {
    jint t = 0;
    //  https://learn.microsoft.com/ja-jp/windows/win32/api/fileapi/nf-fileapi-readfile
    if (!ReadFile (wp->handle, &buf[off + k], len - k, (LPDWORD) &t, NULL)) {
      jint error = GetLastError ();
      if (error == ERROR_BROKEN_PIPE ||  //109 パイプが終了した。誰かが開いて閉じた
          error == ERROR_PIPE_LISTENING || //536 誰も開いていない
          error == ERROR_OPERATION_ABORTED) {  //995 スレッドが終了したか操作が取り消された
        break;
      } else {
        char mes[256];
        sprintf (mes, "ReadFile returned error code %ld", error);
        (*env)->ReleasePrimitiveArrayCritical (env, b, buf, 0);
        throwIOException (env, mes);
        return 0;
      }
    }
    k += t;
  }
  (*env)->ReleasePrimitiveArrayCritical (env, b, buf, 0);
  //入力した長さを返す
  return k;
}

//------------------------------------------------------------------------
//len = readNBytes (b, off, len)
//  配列に入力して入力した長さを返す
//  入力する長さが0のときは何もせず0を返す。さもなくばすべて入力するまでブロックする
//  len  入力した長さ
//  b    配列
//  off  開始位置
//  len  入力する長さ
JNIEXPORT jint JNICALL Java_xeij_NamedPipeInputStream_00024Win_readNBytes (JNIEnv *env, jobject obj, jbyteArray b, jint off, jint len) {
  //変数を取り出す
  jclass cls = (*env)->GetObjectClass (env, obj);
  jfieldID wpid = (*env)->GetFieldID (env, cls, "wp", "J");
  workarea_t *wp = (workarea_t *) (*env)->GetLongField (env, obj, wpid);
  //閉じていたらエラー
  if (wp == NULL || wp->handle == INVALID_HANDLE_VALUE) {
    throwIOException (env, "named pipe not open");
    return 0;
  }
  //開始位置と入力する長さを確認する
  jint cap = (*env)->GetArrayLength (env, b);  //配列の長さ
  if (off < 0 || len < 0 || cap < off + len) {  //範囲外
    char mes[256];
    sprintf (mes, "b.length=%ld, off=%ld, len=%ld", cap, off, len);
    throwIndexOutOfBoundsException (env, mes);
    return 0;
  }
  //入力する長さが0のときは何もせず0を返す
  if (len == 0) {
    return 0;
  }
  //lenバイト入力する
  //  GetPrimitiveArrayCriticalに成功したらReleasePrimitiveArrayCriticalまで他のJNI関数を呼び出してはならない。入れ子にすることはできる
  char *buf = (char *) (*env)->GetPrimitiveArrayCritical (env, b, 0);
  jint k = 0;
  while (k < len) {
    jint t = 0;
    //  https://learn.microsoft.com/ja-jp/windows/win32/api/fileapi/nf-fileapi-readfile
    if (!ReadFile (wp->handle, &buf[off + k], len - k, (LPDWORD) &t, NULL)) {
      jint error = GetLastError ();
      if (error == ERROR_BROKEN_PIPE ||  //109 パイプが終了した。誰かが開いて閉じた
          error == ERROR_PIPE_LISTENING || //536 誰も開いていない
          error == ERROR_OPERATION_ABORTED) {  //995 スレッドが終了したか操作が取り消された
        break;
      } else {
        char mes[256];
        sprintf (mes, "ReadFile returned error code %ld", error);
        (*env)->ReleasePrimitiveArrayCritical (env, b, buf, 0);
        throwIOException (env, mes);
        return 0;
      }
    }
    k += t;
  }
  (*env)->ReleasePrimitiveArrayCritical (env, b, buf, 0);
  //入力した長さを返す
  return k;
}
